"""A small template for ROP-based experiments."""


import argparse
from typing import Final

from moptipy.algorithms.mo.nsga2 import NSGA2
from moptipy.api.experiment import run_experiment
from moptipy.api.mo_execution import MOExecution
from moptipy.mo.problem.weighted_sum import WeightedSum
from moptipy.operators.intspace.op0_random import Op0Random
from moptipy.operators.intspace.op1_mnormal import Op1MNormal
from moptipy.operators.intspace.op2_uniform import Op2Uniform
from moptipy.spaces.intspace import IntSpace
from pycommons.io.console import logger
from pycommons.io.path import Path
from pycommons.types import check_int_range

from moptipyapps.prodsched.instance import Instance
from moptipyapps.prodsched.instances import get_instances
from moptipyapps.prodsched.multistatistics import MultiStatisticsSpace
from moptipyapps.prodsched.objectives.max_stocklevel import MaxStockLevel
from moptipyapps.prodsched.objectives.worst_and_mean_fill_rate import (
    WorstAndMeanFillRate,
)
from moptipyapps.prodsched.rop_multisimulation import ROPMultiSimulation
from moptipyapps.utils.shared import moptipyapps_argparser


def run(dest: str, instances: str, n_inst: int, n_runs: int,
        max_fes: int, ps: int) -> None:
    """
    Run the experiment.

    :param dest: the destination directory
    :param instances: the directory with the instances
    :param n_inst: the number of instances
    :param n_runs: the number of runs
    :param max_fes: the maximum FEs
    :param ps: the population size
    """
    logger(f"Beginning experiment with dest={dest!r}, instances={instances!r}"
           f", n_inst={n_inst}, n_runs={n_runs}, and max_fes={max_fes}.")
    use_dest: Final[Path] = Path(dest)
    use_dest.ensure_dir_exists()
    logger(f"Destination folder is {use_dest!r}.")

    use_insts: Final[Path] = Path(instances)
    use_insts.ensure_dir_exists()
    logger(f"Instances folder is {use_insts!r}.")

    check_int_range(n_inst, "n_inst", 1, 128)
    check_int_range(max_fes, "max_fes", 10, 10 ** 10)
    check_int_range(ps, "ps", 4, 16384)

    logger(f"Loading {n_inst} instances from {use_insts!r}.")
    insts: Final[tuple[Instance, ...]] = get_instances(n_inst, instances)
    if tuple.__len__(insts) != n_inst:
        raise ValueError("Could not load required instances.")
    logger(f"Loaded {n_inst} instances from {use_insts!r}.")

    n_prod: int | None = None
    for inst in insts:
        if n_prod is None:
            n_prod = inst.n_products
        elif n_prod != inst.n_products:
            raise ValueError("Inconsistent number of products!")
    if n_prod is None:
        raise ValueError("No instances?")

    search_space: Final[IntSpace] = IntSpace(n_prod, 0, 63)
    op0: Final[Op0Random] = Op0Random(search_space)
    op1: Final[Op1MNormal] = Op1MNormal(search_space, sd=2.5)
    op2: Final = Op2Uniform()
    algo: Final[NSGA2] = NSGA2(op0, op1, op2, ps, 1 / min(16, ps))
    encoding: Final[ROPMultiSimulation] = ROPMultiSimulation(insts)
    f1: Final[WorstAndMeanFillRate] = WorstAndMeanFillRate()
    f2: Final[MaxStockLevel] = MaxStockLevel()
    ws: Final[WeightedSum] = WeightedSum((f1, f2), (
        (1 / (f1.upper_bound() - f1.lower_bound())), 1 / (2 * n_prod)))
    solution_space: Final[MultiStatisticsSpace] = MultiStatisticsSpace(insts)

    def __setup(_) -> MOExecution:
        """
        Set up the experiment.

        :return: the execution
        """
        return (MOExecution()
                .set_search_space(search_space)
                .set_algorithm(algo)
                .set_solution_space(solution_space)
                .set_objective(ws)
                .set_encoding(encoding)
                .set_max_fes(max_fes)
                .set_log_improvements(True))

    run_experiment(base_dir=use_dest, instances=(lambda: "all", ),
                   setups=(__setup, ), n_runs=n_runs,
                   pre_warmup_fes=2, perform_warmup=False,
                   perform_pre_warmup=True)


# Run the experiment from the command line
if __name__ == "__main__":
    parser: Final[argparse.ArgumentParser] = moptipyapps_argparser(
        __file__, "ROP-based MFC Optimization",
        "Run a small experiment with ROP-based MFC optimization.")
    parser.add_argument(
        "dest", help="the directory to store the experimental results under",
        type=Path, nargs="?", default="./results/")
    parser.add_argument(
        "insts", help="the directory with the instances",
        type=Path, nargs="?", default="./instances/")
    parser.add_argument(
        "n_inst", help="the number of instances",
        type=int, nargs="?", default=11)
    parser.add_argument(
        "n_runs", help="the number of runs",
        type=int, nargs="?", default=31)
    parser.add_argument(
        "max_fes", help="the number of FEs per run",
        type=int, nargs="?", default=8192)
    parser.add_argument(
        "ps", help="the population size of NSGA-II",
        type=int, nargs="?", default=64)
    args: Final[argparse.Namespace] = parser.parse_args()
    run(args.dest, args.insts, args.n_inst, args.n_runs, args.max_fes,
        args.ps)



"""Generate instances for training and testing."""

from typing import Final

from moptipy.utils.nputils import rand_seeds_from_str
from pycommons.io.path import Path
from pycommons.types import check_int_range

from moptipyapps.prodsched.instance import (
    Instance,
    instance_sort_key,
    load_instances,
    store_instances,
)
from moptipyapps.prodsched.mfc_generator import (
    sample_mfc_instance,
)


def get_instances(n: int, inst_dir: str) -> tuple[Instance, ...]:
    """
    Get the instances for the experiment.

    :param n: the expected number of instances
    :param inst_dir: the instance directory
    """
    check_int_range(n, "n", 1, 1_000_000)
    use_dir: Final[Path] = Path(inst_dir)

    has_instances: bool = False
    if use_dir.exists():
        has_instances = True
        try:
            next(use_dir.list_dir(True, False))
        except StopIteration:
            has_instances = False

    if has_instances:
        loaded: Final[tuple[Instance, ...]] = load_instances(use_dir)
        loaded_len: Final[int] = tuple.__len__(loaded)
        if loaded_len != n:
            raise ValueError(f"Excepted {n} instances, got {loaded_len}.")
        return loaded

    insts: Final[list[Instance]] = []
    for seed in rand_seeds_from_str("mfc", n):
        insts.append(sample_mfc_instance(seed=seed))

    insts.sort(key=instance_sort_key)
    store_instances(use_dir, insts)
    return tuple(insts)



"""A statistics record for the simulation."""

from itertools import chain
from typing import Callable, Final, Generator

from moptipy.utils.logger import KEY_VALUE_SEPARATOR
from pycommons.io.csv import CSV_SEPARATOR, SCOPE_SEPARATOR
from pycommons.math.stream_statistics import (
    KEY_MAXIMUM,
    KEY_MEAN_ARITH,
    KEY_MINIMUM,
    KEY_STDDEV,
    StreamStatistics,
)
from pycommons.strings.string_conv import num_or_none_to_str
from pycommons.types import check_int_range, type_error

#: the name of the statistics key
COL_STAT: Final[str] = "stat"
#: the total column name
COL_TOTAL: Final[str] = "total"
#: the statistics rate
KEY_RATE: Final[str] = "rate"
#: the product column prefix
COL_PRODUCT_PREFIX: Final[str] = "product_"
#: the mean TRP row
ROW_TRP: Final[str] = "trp"
#: the fill rate row
ROW_FILL_RATE: Final[str] = f"fill{SCOPE_SEPARATOR}{KEY_RATE}"
#: the CWT row
ROW_CWT: Final[str] = "cwt"
#: the mean stock level row
ROW_STOCK_LEVEL_MEAN: Final[str] = \
    f"stocklevel{SCOPE_SEPARATOR}{KEY_MEAN_ARITH}"
#: the fulfilled rate
ROW_FULFILLED_RATE: Final[str] = f"fulfilled{SCOPE_SEPARATOR}{KEY_RATE}"
#: the simulation time getter
ROW_SIMULATION_TIME: Final[str] = \
    f"time{SCOPE_SEPARATOR}s{KEY_VALUE_SEPARATOR}"

#: the statistics that we will print
__STATS: tuple[tuple[str, Callable[[
    StreamStatistics], int | float | None]], ...] = (
    (KEY_MINIMUM, StreamStatistics.getter_or_none(KEY_MINIMUM)),
    (KEY_MEAN_ARITH, StreamStatistics.getter_or_none(KEY_MEAN_ARITH)),
    (KEY_MAXIMUM, StreamStatistics.getter_or_none(KEY_MAXIMUM)),
    (KEY_STDDEV, StreamStatistics.getter_or_none(KEY_STDDEV)))


class Statistics:
    """A statistics record based on production scheduling."""

    def __init__(self, n_products: int) -> None:
        """
        Create the statistics record.

        :param n_products: the number of products
        """
        check_int_range(n_products, "n_products", 1, 1_000_000_000)
        #: the production time statistics per-product
        self.production_times: Final[list[
            StreamStatistics | None]] = [None] * n_products
        #: the overall production time statistics
        self.production_time: StreamStatistics | None = None
        #: the fraction of demands that were immediately satisfied,
        #: on a per-product basis
        self.immediate_rates: Final[list[int | float | None]] = (
            [None] * n_products)
        #: the overall fraction of immediately satisfied demands
        self.immediate_rate: int | float | None = None
        #: the average waiting time for all demands that were not immediately
        #: satisfied -- only counting demands that were actually satisfied
        self.waiting_times: Final[list[
            StreamStatistics | None]] = [None] * n_products
        #: the overall waiting time for all demands that were not immediately
        #: satisfied -- only counting demands that were actually satisfied
        self.waiting_time: StreamStatistics | None = None
        #: the fraction of demands that were fulfilled, on a per-product basis
        self.fulfilled_rates: Final[list[
            int | float | None]] = [None] * n_products
        #: the fraction of demands that were fulfilled overall
        self.fulfilled_rate: int | float | None = None
        #: the average stock level, on a per-product basis
        self.stock_levels: Final[list[
            int | float | None]] = [None] * n_products
        #: the overall average stock level
        self.stock_level: int | float | None = None
        #: the nano seconds used by the simulation
        self.simulation_time_nanos: int | float | None = None

    def __str__(self) -> str:
        """Convert this object to a string."""
        return "\n".join(to_stream(self))

    def copy_from(self, stat: "Statistics") -> None:
        """
        Copy the contents of another statistics record.

        :param stat: the other statistics record
        """
        if not isinstance(stat, Statistics):
            raise type_error(stat, "stat", Statistics)
        self.production_times[:] = stat.production_times
        self.production_time = stat.production_time
        self.immediate_rates[:] = stat.immediate_rates
        self.immediate_rate = stat.immediate_rate
        self.waiting_times[:] = stat.waiting_times
        self.waiting_time = stat.waiting_time
        self.fulfilled_rates[:] = stat.fulfilled_rates
        self.fulfilled_rate = stat.fulfilled_rate
        self.stock_levels[:] = stat.stock_levels
        self.stock_level = stat.stock_level
        self.simulation_time_nanos = stat.simulation_time_nanos


def to_stream(stats: Statistics) -> Generator[str, None, None]:
    """
    Write a statistics record to a stream.

    :param stats: the statistics record
    :return: the stream of data
    """
    n_products: Final[int] = list.__len__(stats.production_times)
    nts: Final[Callable[[int | float | None], str]] = num_or_none_to_str

    yield str.join(CSV_SEPARATOR, chain((
        COL_STAT, COL_TOTAL), (f"{COL_PRODUCT_PREFIX}{i}" for i in range(
            n_products))))

    for key, alle, single in (
            (ROW_TRP, stats.production_times, stats.production_time),
            (ROW_CWT, stats.waiting_times, stats.waiting_time)):
        for stat, call in __STATS:
            yield str.join(CSV_SEPARATOR, chain((
                f"{key}{SCOPE_SEPARATOR}{stat}", nts(call(single))), (
                map(nts, map(call, alle)))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_FILL_RATE, nts(stats.immediate_rate)), (
        map(nts, stats.immediate_rates))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_STOCK_LEVEL_MEAN, nts(stats.stock_level)), (
        map(nts, stats.stock_levels))))
    yield str.join(CSV_SEPARATOR, chain((
        ROW_FULFILLED_RATE, nts(stats.fulfilled_rate)), (
        map(nts, stats.fulfilled_rates))))
    yield f"{ROW_SIMULATION_TIME}{nts(
        stats.simulation_time_nanos / 1_000_000_000)}"



"""An objective function for minimizing the maximal stocklevel."""

from math import inf

from moptipy.api.objective import Objective

from moptipyapps.prodsched.multistatistics import MultiStatistics


class MaxStockLevel(Objective):
    """Compute the worst stock level."""

    def evaluate(self, x: MultiStatistics) -> int | float:
        """
        Get the worst (largest) stocklevel.

        :param x: the multi-statistics
        :return: the worst stock level
        """
        max_sl: int | float = 0
        for stat in x.per_instance:
            sl = stat.stock_level
            if (sl is None) or not (0 <= sl <= 1_000_000_000):
                return inf
            max_sl = max(max_sl, sl)
        return max_sl

    def lower_bound(self) -> int:
        """
        Get the lower bound of the maximum storage level.

        :retval 0: always
        """
        return 0

    def __str__(self) -> str:
        """
        Get the name of the objective function.

        :return: `maxStorageLevel`
        :retval "maxStorageLevel": always
        """
        return "maxStorageLevel"



"""Maximize the worst and average immediate rates."""


from moptipy.api.objective import Objective

from moptipyapps.prodsched.multistatistics import MultiStatistics


class WorstAndMeanFillRate(Objective):
    """Compute the worst immediate rate and return `1 -` of it."""

    def evaluate(self, x: MultiStatistics) -> int | float:
        """
        Get the negated worst immediate rate.

        :param x: the multi-statistics
        :return: the worst stock level
        """
        min_imm: int | float = 1
        avg_imm: int | float = 1
        for stat in x.per_instance:
            for sl in stat.immediate_rates:
                min_imm = 0 if sl is None else min(min_imm, sl)
            avg_imm = 0 if stat.immediate_rate is None else min(
                avg_imm, stat.immediate_rate)
        return (1 - min_imm) * 100 + (1 - avg_imm)

    def lower_bound(self) -> int:
        """
        Get the lower bound of the inverted minimum immediate rate.

        :retval 0: always
        """
        return 0

    def upper_bound(self) -> int:
        """
        Get the upper bound of the inverted minimum immediate rate.

        :retval 1: always
        """
        return 101

    def __str__(self) -> str:
        """
        Get the name of the objective function.

        :return: `worstMinAndMeanFillRate`
        :retval "worstMinAndMeanFillRate": always
        """
        return "worstAndMeanFillRate"



"""A statistics record for the simulation."""

from dataclasses import dataclass
from typing import Final, Generator, Iterable

from moptipy.api.space import Space
from moptipy.utils.logger import (
    KeyValueLogSection,
)
from pycommons.types import type_error

from moptipyapps.prodsched.instance import Instance
from moptipyapps.prodsched.statistics import Statistics
from moptipyapps.prodsched.statistics import to_stream as stat_to_stream


@dataclass(order=False, frozen=True)
class MultiStatistics:
    """A set of statistics gathered over multiple instances."""

    #: the per-instance statistics
    per_instance: tuple[Statistics, ...]
    #: the instance names
    inst_names: tuple[str, ...]

    def __init__(self, instances: Iterable[Instance]) -> None:
        """
        Create the multi-statistics object.

        :param instances: the instances for which we create the statistics
        """
        object.__setattr__(self, "per_instance", tuple(
            Statistics(inst.n_products) for inst in instances))
        object.__setattr__(self, "inst_names", tuple(
            inst.name for inst in instances))


def to_stream(multi: MultiStatistics) -> Generator[str, None, None]:
    """
    Convert a multi-statistics object to a stream.

    :param multi: the multi-statistics object
    :return: the stream of strings
    """
    if not isinstance(multi, MultiStatistics):
        raise type_error(multi, "multi", MultiStatistics)
    for i, ss in enumerate(multi.per_instance):
        yield f"-------- Instance {i}: {multi.inst_names[i]!r} -------"
        yield from stat_to_stream(ss)


class MultiStatisticsSpace(Space):
    """An implementation of the `Space` API of for multiple statistics."""

    def __init__(self, instances: tuple[Instance, ...]) -> None:
        """
        Create a multi-statistics space.

        :param instances: the instances
        """
        if not isinstance(instances, tuple):
            raise type_error(instances, "instances", tuple)
        for inst in instances:
            if not isinstance(inst, Instance):
                raise type_error(inst, "instance", Instance)
        #: The instance to which the packings apply.
        self.instances: Final[tuple[Instance, ...]] = instances

    def copy(self, dest: MultiStatistics, source: MultiStatistics) -> None:
        """
        Copy one multi-statistics to another one.

        :param dest: the destination multi-statistics
        :param source: the source multi-statistics
        """
        for i, d in enumerate(dest.per_instance):
            d.copy_from(source.per_instance[i])

    def create(self) -> MultiStatistics:
        """
        Create an empty multi-statistics record.

        :return: the empty multi-statistics record
        """
        return MultiStatistics(self.instances)

    def to_str(self, x: MultiStatistics) -> str:
        """
        Convert a multi-statistics to a string.

        :param x: the packing
        :return: a string corresponding to the multi-statistics
        """
        return "\n".join(to_stream(x))

    def is_equal(self, x1: MultiStatistics, x2: MultiStatistics) -> bool:
        """
        Check if two multi-statistics have the same contents.

        :param x1: the first multi-statistics
        :param x2: the second multi-statistics
        :return: `True` if both multi-statistics have the same content
        """
        return (x1 is x2) or (x1.per_instance == x2.per_instance)

    def from_str(self, text: str) -> MultiStatistics:
        """
        Convert a string to a multi-statistics.

        :param text: the string
        :return: the multi-statistics
        """
        if not isinstance(text, str):
            raise type_error(text, "text", str)
        raise NotImplementedError

    def validate(self, x: MultiStatistics) -> None:
        """
        Check if a multi-statistics is valid.

        :param x: the multi-statistics
        :raises TypeError: if any component of the multi-statistics is of the
            wrong type
        :raises ValueError: if the multi-statistics is not feasible
        """
        if not isinstance(x, MultiStatistics):
            raise type_error(x, "x", MultiStatistics)
        if not isinstance(x.per_instance, tuple):
            raise type_error(x.per_instance, "x.per_instance", tuple)
        for s in x.per_instance:
            if not isinstance(s, Statistics):
                raise type_error(s, "x.per_instance[i]", Statistics)

    def n_points(self) -> int:
        """
        Get the number of possible multi-statistics.

        :return: just some arbitrary number
        """
        return 100 ** tuple.__len__(self.instances)

    def __str__(self) -> str:
        """
        Get the name of the multi-statistics space.

        :return: the name
        """
        return f"multistats_{tuple.__len__(self.instances)}"

    def log_parameters_to(self, logger: KeyValueLogSection) -> None:
        """
        Log the parameters of the space to the given logger.

        :param logger: the logger for the parameters
        """
        super().log_parameters_to(logger)
        for i, inst in enumerate(self.instances):
            logger.key_value(f"inst_{i}", inst.name)



"""
Different encodings for the two-dimensional bin packing problem.

The following encodings are implemented:

- The improved bottom-left encoding 1,
  :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_1`, processes a
  permutation of objects from beginning to end and places the objects into
  the last bin (according to the improved-bottom-left method) until that bin
  is full and then begins to put them into the next bin.
- The improved bottom-left encoding 2,
  :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_2`, processes a
  permutation of objects from beginning to end and places the objects into
  the first bin into which they fit (according to the improved-bottom-left
  method). It is slower than
  :mod:`~moptipyapps.binpacking2d.encodings.ibl_encoding_1` but its results
  are never worse for any permutation and better for several.

Important initial work on this code has been contributed by Mr. Rui ZHAO
(赵睿), <zr1329142665@163.com> a Master's student at the Institute of Applied
Optimization (应用优化研究所) of the School of
Artificial Intelligence and Big Data (人工智能与大数据学院) at Hefei University
(合肥大学) in Hefei, Anhui, China (中国安徽省合肥市) under the supervision of
Prof. Dr. Thomas Weise (汤卫思教授).
"""



"""An objective function for maximizing the worst immediate rate."""


from moptipy.api.objective import Objective

from moptipyapps.prodsched.multistatistics import MultiStatistics


class WorstFillRate(Objective):
    """Compute the worst immediate rate and return `1 -` of it."""

    def evaluate(self, x: MultiStatistics) -> int | float:
        """
        Get the negated worst immediate rate.

        :param x: the multi-statistics
        :return: the worst stock level
        """
        min_imm: int | float = 1
        for stat in x.per_instance:
            for sl in stat.immediate_rates:
                if (sl is None) or not (0 < sl <= 1):
                    return 1
                min_imm = min(min_imm, sl)
        return 1 - min_imm

    def lower_bound(self) -> int:
        """
        Get the lower bound of the inverted minimum immediate rate.

        :retval 0: always
        """
        return 0

    def upper_bound(self) -> int:
        """
        Get the upper bound of the inverted minimum immediate rate.

        :retval 1: always
        """
        return 1

    def __str__(self) -> str:
        """
        Get the name of the objective function.

        :return: `worstFillRate`
        :retval "worstFillRate": always
        """
        return "worstFillRate"



"""
A re-order-point-based simulation.

>>> from moptipyapps.prodsched.simulation import PrintingListener
>>> instance = Instance(
...     name="test1", n_products=2, n_customers=4, n_stations=2, n_demands=4,
...     time_end_warmup=3000, time_end_measure=10000,
...     routes=[[0, 1], [1, 0]],
...     demands=[Demand(arrival=100, deadline=100, demand_id=0,
...                     customer_id=0, product_id=0, amount=1, measure=False),
...              Demand(arrival=3100, deadline=3100, demand_id=1,
...                     customer_id=1, product_id=0, amount=1, measure=True),
...              Demand(arrival=500, deadline=500, demand_id=2,
...                     customer_id=2, product_id=1, amount=1, measure=False),
...              Demand(arrival=4000, deadline=5000, demand_id=3,
...                     customer_id=3, product_id=1, amount=1, measure=True)],
...     warehous_at_t0=[0, 0],
...     station_product_unit_times=[[[10.0, 50.0, 15.0, 100.0],
...                                  [ 5.0, 20.0,  7.0,  35.0, 4.0, 50.0]],
...                                 [[ 5.0, 24.0,  7.0,  80.0],
...                                  [ 3.0, 21.0,  6.0,  50.0,]]])
>>> rop_sim = ROPSimulation(instance, PrintingListener(print_time=False))
>>> rop_sim.set_rop((10, 20))
>>> rop_sim.ctrl_run()
start
T=0.0: product=0, amount=0, in_warehouse=0, in_production=0, 0 pending demands
T=0.0: station=0, 1 jobs queued
T=0.0: start j(id: 0, p: 0, am: 11, ar: 0, me: F, c: F, st: 0, sp: 0)\
 at station 0
T=0.0: product=1, amount=0, in_warehouse=0, in_production=0, 0 pending demands
T=0.0: station=1, 1 jobs queued
T=0.0: start j(id: 1, p: 1, am: 21, ar: 0, me: F, c: F, st: 0, sp: 0)\
 at station 1
T=84.0: finished j(id: 1, p: 1, am: 21, ar: 0, me: F, c: F, st: 0, sp: 0)\
 at station 1
T=100.0: product=0, amount=0, in_warehouse=0, in_production=11,\
 1 pending demands
T=130.0: finished j(id: 0, p: 0, am: 11, ar: 0, me: F, c: F, st: 0,\
 sp: 0) at station 0
T=130.0: station=0, 2 jobs queued
T=130.0: start j(id: 1, p: 1, am: 21, ar: 0, me: F, c: F, st: 84,\
 sp: 1) at station 0
T=130.0: station=1, 1 jobs queued
T=130.0: start j(id: 0, p: 0, am: 11, ar: 0, me: F, c: F, st: 130,\
 sp: 1) at station 1
T=197.0: finished j(id: 0, p: 0, am: 11, ar: 0, me: F, c: T, st: 130,\
 sp: 1) at station 1
T=197.0: product=0, amount=11, in_warehouse=0, in_production=1,\
 1 pending demands
T=197.0: d(id: 0, p: 0, c: 0, am: 1, ar: 100, dl: 100, me: F) statisfied
T=197.0: 10 units of product 0 in warehouse
T=240.0: finished j(id: 1, p: 1, am: 21, ar: 0, me: F, c: T, st: 84,\
 sp: 1) at station 0
T=240.0: station=0, 1 jobs queued
T=240.0: start j(id: 2, p: 0, am: 1, ar: 100, me: F, c: F, st: 100,\
 sp: 0) at station 0
T=240.0: product=1, amount=21, in_warehouse=0, in_production=0,\
 0 pending demands
T=240.0: 21 units of product 1 in warehouse
T=250.0: finished j(id: 2, p: 0, am: 1, ar: 100, me: F, c: F, st: 100,\
 sp: 0) at station 0
T=250.0: station=1, 1 jobs queued
T=250.0: start j(id: 2, p: 0, am: 1, ar: 100, me: F, c: F, st: 250, sp: 1)\
 at station 1
T=255.0: finished j(id: 2, p: 0, am: 1, ar: 100, me: F, c: T, st: 250, sp: 1)\
 at station 1
T=255.0: product=0, amount=1, in_warehouse=10, in_production=0,\
 0 pending demands
T=255.0: 11 units of product 0 in warehouse
T=500.0: product=1, amount=0, in_warehouse=21, in_production=0,\
 1 pending demands
T=500.0: d(id: 2, p: 1, c: 2, am: 1, ar: 500, dl: 500, me: F) statisfied
T=500.0: 20 units of product 1 in warehouse
T=500.0: station=1, 1 jobs queued
T=500.0: start j(id: 3, p: 1, am: 1, ar: 500, me: F, c: F, st: 500,\
 sp: 0) at station 1
T=503.0: finished j(id: 3, p: 1, am: 1, ar: 500, me: F, c: F, st: 500,\
 sp: 0) at station 1
T=503.0: station=0, 1 jobs queued
T=503.0: start j(id: 3, p: 1, am: 1, ar: 500, me: F, c: F, st: 503,\
 sp: 1) at station 0
T=508.0: finished j(id: 3, p: 1, am: 1, ar: 500, me: F, c: T, st: 503,\
 sp: 1) at station 0
T=508.0: product=1, amount=1, in_warehouse=20, in_production=0,\
 0 pending demands
T=508.0: 21 units of product 1 in warehouse
T=3100.0! product=0, amount=0, in_warehouse=11, in_production=0,\
 1 pending demands
T=3100.0! d(id: 1, p: 0, c: 1, am: 1, ar: 3100, dl: 3100, me: T) statisfied
T=3100.0! 10 units of product 0 in warehouse
T=3100.0! station=0, 1 jobs queued
T=3100.0! start j(id: 4, p: 0, am: 1, ar: 3100, me: T, c: F, st: 3100,\
 sp: 0) at station 0
T=3110.0! finished j(id: 4, p: 0, am: 1, ar: 3100, me: T, c: F, st: 3100,\
 sp: 0) at station 0
T=3110.0! station=1, 1 jobs queued
T=3110.0! start j(id: 4, p: 0, am: 1, ar: 3100, me: T, c: F, st: 3110,\
 sp: 1) at station 1
T=3117.0! finished j(id: 4, p: 0, am: 1, ar: 3100, me: T, c: T, st: 3110,\
 sp: 1) at station 1
T=3117.0! product=0, amount=1, in_warehouse=10, in_production=0,\
 0 pending demands
T=3117.0! 11 units of product 0 in warehouse
T=4000.0! product=1, amount=0, in_warehouse=21, in_production=0,\
 1 pending demands
T=4000.0! d(id: 3, p: 1, c: 3, am: 1, ar: 4000, dl: 5000, me: T) statisfied
T=4000.0! 20 units of product 1 in warehouse
T=4000.0! station=1, 1 jobs queued
T=4000.0! start j(id: 5, p: 1, am: 1, ar: 4000, me: T, c: F, st: 4000,\
 sp: 0) at station 1
T=4003.0! finished j(id: 5, p: 1, am: 1, ar: 4000, me: T, c: F, st: 4000,\
 sp: 0) at station 1
T=4003.0! station=0, 1 jobs queued
T=4003.0! start j(id: 5, p: 1, am: 1, ar: 4000, me: T, c: F, st: 4003,\
 sp: 1) at station 0
T=4008.0! finished j(id: 5, p: 1, am: 1, ar: 4000, me: T, c: T, st: 4003,\
 sp: 1) at station 0
T=4008.0! product=1, amount=1, in_warehouse=20, in_production=0,\
 0 pending demands
T=4008.0! 21 units of product 1 in warehouse
T=4008.0 -- finished

>>> instance = Instance(
...     name="test2", n_products=2, n_customers=1, n_stations=2, n_demands=5,
...     time_end_warmup=21, time_end_measure=10000,
...     routes=[[0, 1], [1, 0]],
...     demands=[[0, 0, 1, 4, 20, 90], [1, 0, 0, 5, 22, 200],
...              [2, 0, 1, 4, 30, 200], [3, 0, 1, 6, 60, 200],
...              [4, 0, 0, 3, 5, 2000]],
...     warehous_at_t0=[2, 1],
...     station_product_unit_times=[[[10.0, 50.0, 15.0, 100.0],
...                                  [ 5.0, 20.0,  7.0,  35.0, 4.0, 50.0]],
...                                 [[ 5.0, 24.0,  7.0,  80.0],
...                                  [ 3.0, 21.0,  6.0,  50.0,]]])
>>> rop_sim = ROPSimulation(instance, PrintingListener(print_time=False))
>>> rop_sim.set_rop((10, 10))
>>> rop_sim.ctrl_run()
start
T=0.0: product=0, amount=2, in_warehouse=0, in_production=0, 0 pending demands
T=0.0: 2 units of product 0 in warehouse
T=0.0: station=0, 1 jobs queued
T=0.0: start j(id: 0, p: 0, am: 9, ar: 0, me: F, c: F, st: 0, sp: 0)\
 at station 0
T=0.0: product=1, amount=1, in_warehouse=0, in_production=0, 0 pending demands
T=0.0: 1 units of product 1 in warehouse
T=0.0: station=1, 1 jobs queued
T=0.0: start j(id: 1, p: 1, am: 10, ar: 0, me: F, c: F, st: 0, sp: 0)\
 at station 1
T=5.0: product=0, amount=0, in_warehouse=2, in_production=9, 1 pending demands
T=20.0: product=1, amount=0, in_warehouse=1, in_production=10,\
 1 pending demands
T=22.0! product=0, amount=0, in_warehouse=2, in_production=12,\
 2 pending demands
T=30.0! product=1, amount=0, in_warehouse=1, in_production=14,\
 2 pending demands
T=39.0: finished j(id: 1, p: 1, am: 10, ar: 0, me: F, c: F, st: 0, sp: 0)\
 at station 1
T=39.0! station=1, 2 jobs queued
T=39.0: start j(id: 3, p: 1, am: 4, ar: 20, me: F, c: F, st: 20, sp: 0)\
 at station 1
T=57.0: finished j(id: 3, p: 1, am: 4, ar: 20, me: F, c: F, st: 20, sp: 0)\
 at station 1
T=57.0! station=1, 1 jobs queued
T=57.0! start j(id: 5, p: 1, am: 4, ar: 30, me: T, c: F, st: 30, sp: 0)\
 at station 1
T=60.0! product=1, amount=0, in_warehouse=1, in_production=18,\
 3 pending demands
T=69.0! finished j(id: 5, p: 1, am: 4, ar: 30, me: T, c: F, st: 30, sp: 0)\
 at station 1
T=69.0! station=1, 1 jobs queued
T=69.0! start j(id: 6, p: 1, am: 6, ar: 60, me: T, c: F, st: 60, sp: 0)\
 at station 1
T=102.0! finished j(id: 6, p: 1, am: 6, ar: 60, me: T, c: F, st: 60, sp: 0)\
 at station 1
T=110.0: finished j(id: 0, p: 0, am: 9, ar: 0, me: F, c: F, st: 0, sp: 0)\
 at station 0
T=110.0! station=0, 6 jobs queued
T=110.0: start j(id: 2, p: 0, am: 3, ar: 5, me: F, c: F, st: 5, sp: 0)\
 at station 0
T=110.0! station=1, 1 jobs queued
T=110.0: start j(id: 0, p: 0, am: 9, ar: 0, me: F, c: F, st: 110, sp: 1)\
 at station 1
T=140.0: finished j(id: 2, p: 0, am: 3, ar: 5, me: F, c: F, st: 5, sp: 0)\
 at station 0
T=140.0! station=0, 5 jobs queued
T=140.0! start j(id: 4, p: 0, am: 5, ar: 22, me: T, c: F, st: 22, sp: 0)\
 at station 0
T=171.0: finished j(id: 0, p: 0, am: 9, ar: 0, me: F, c: T, st: 110, sp: 1)\
 at station 1
T=171.0! station=1, 1 jobs queued
T=171.0: start j(id: 2, p: 0, am: 3, ar: 5, me: F, c: F, st: 140, sp: 1)\
 at station 1
T=171.0! product=0, amount=9, in_warehouse=2, in_production=8,\
 2 pending demands
T=171.0: d(id: 4, p: 0, c: 0, am: 3, ar: 5, dl: 2000, me: F) statisfied
T=171.0! d(id: 1, p: 0, c: 0, am: 5, ar: 22, dl: 200, me: T) statisfied
T=171.0! 3 units of product 0 in warehouse
T=186.0: finished j(id: 2, p: 0, am: 3, ar: 5, me: F, c: T, st: 140, sp: 1)\
 at station 1
T=186.0! product=0, amount=3, in_warehouse=3, in_production=5,\
 0 pending demands
T=186.0! 6 units of product 0 in warehouse
T=210.0! finished j(id: 4, p: 0, am: 5, ar: 22, me: T, c: F, st: 22, sp: 0)\
 at station 0
T=210.0! station=0, 4 jobs queued
T=210.0: start j(id: 1, p: 1, am: 10, ar: 0, me: F, c: F, st: 39, sp: 1)\
 at station 0
T=210.0! station=1, 1 jobs queued
T=210.0! start j(id: 4, p: 0, am: 5, ar: 22, me: T, c: F, st: 210, sp: 1)\
 at station 1
T=245.0! finished j(id: 4, p: 0, am: 5, ar: 22, me: T, c: T, st: 210, sp: 1)\
 at station 1
T=245.0! product=0, amount=5, in_warehouse=6, in_production=0,\
 0 pending demands
T=245.0! 11 units of product 0 in warehouse
T=263.0: finished j(id: 1, p: 1, am: 10, ar: 0, me: F, c: T, st: 39, sp: 1)\
 at station 0
T=263.0! station=0, 3 jobs queued
T=263.0: start j(id: 3, p: 1, am: 4, ar: 20, me: F, c: F, st: 57, sp: 1)\
 at station 0
T=263.0! product=1, amount=10, in_warehouse=1, in_production=14,\
 3 pending demands
T=263.0: d(id: 0, p: 1, c: 0, am: 4, ar: 20, dl: 90, me: F) statisfied
T=263.0! d(id: 2, p: 1, c: 0, am: 4, ar: 30, dl: 200, me: T) statisfied
T=263.0! 3 units of product 1 in warehouse
T=287.0: finished j(id: 3, p: 1, am: 4, ar: 20, me: F, c: T, st: 57, sp: 1)\
 at station 0
T=287.0! station=0, 2 jobs queued
T=287.0! start j(id: 5, p: 1, am: 4, ar: 30, me: T, c: F, st: 69, sp: 1)\
 at station 0
T=287.0! product=1, amount=4, in_warehouse=3, in_production=10,\
 1 pending demands
T=287.0! d(id: 3, p: 1, c: 0, am: 6, ar: 60, dl: 200, me: T) statisfied
T=287.0! 1 units of product 1 in warehouse
T=303.0! finished j(id: 5, p: 1, am: 4, ar: 30, me: T, c: T, st: 69, sp: 1)\
 at station 0
T=303.0! station=0, 1 jobs queued
T=303.0! start j(id: 6, p: 1, am: 6, ar: 60, me: T, c: F, st: 102, sp: 1)\
 at station 0
T=303.0! product=1, amount=4, in_warehouse=1, in_production=6,\
 0 pending demands
T=303.0! 5 units of product 1 in warehouse
T=337.0! finished j(id: 6, p: 1, am: 6, ar: 60, me: T, c: T, st: 102, sp: 1)\
 at station 0
T=337.0! product=1, amount=6, in_warehouse=5, in_production=0,\
 0 pending demands
T=337.0! 11 units of product 1 in warehouse
T=337.0 -- finished
"""

from typing import Final, Iterable

import numpy as np
from pycommons.types import check_int_range, type_error

from moptipyapps.prodsched.instance import (
    Demand,
    Instance,
)
from moptipyapps.prodsched.simulation import Listener, Simulation


class ROPSimulation(Simulation):
    """Create the re-order point-based simulation."""

    def __init__(self, instance: Instance, listener: Listener) -> None:
        """
        Initialize this simulation.

        :param instance: the instance
        :param listener: the listener
        """
        super().__init__(instance, listener)
        #: the re-order point
        self.__rop: Final[list[int]] = [0] * instance.n_products

    def set_rop(self, rop: Iterable[int] | np.ndarray) -> None:
        """
        Set the re-order point.

        :param rop: the re-order point
        """
        if isinstance(rop, np.ndarray):
            rop = map(int, rop)  # type: ignore
        if not isinstance(rop, Iterable):
            raise type_error(rop, "rop", Iterable)
        i: int = -1
        for i, rv in enumerate(rop):
            self.__rop[i] = check_int_range(
                rv, "rop-value", 0, 1_000_000_000)
        i += 1
        n_prod: Final[int] = list.__len__(self.__rop)
        if i != n_prod:
            raise ValueError(f"Invalid rop length {i}, must be {n_prod}!")

    def event_product(self, time: float,  # pylint: disable=W0613,R0913,R0917
                      product_id: int, amount: int,
                      in_warehouse: int,
                      in_production: int,  # pylint: disable=W0613
                      pending_demands: tuple[Demand, ...]) -> None:
        """Perform a simulation step based on a product."""
        # satisfy demands
        available: int = in_warehouse + amount
        unsatisfied: int = 0
        for demand in pending_demands:
            request: int = demand.amount
            if request <= available:
                self.act_demand_satisfied(demand)
                available -= request
            else:
                unsatisfied += request

        # update warehouse
        if available < in_warehouse:
            self.act_take_from_warehouse(product_id, in_warehouse - available)
        elif available > in_warehouse:
            self.act_store_in_warehouse(product_id, available - in_warehouse)

        # check if re-order point needs to be triggered
        rop: Final[int] = self.__rop[product_id]
        available = available - unsatisfied + in_production
        if available <= rop:
            self.act_produce(product_id, rop - available + 1)



"""A simulator for multiple instances."""
from typing import Final, Iterable

import numpy as np
from moptipy.api.encoding import Encoding
from pycommons.types import type_error

from moptipyapps.prodsched.instance import (
    Instance,
)
from moptipyapps.prodsched.multistatistics import MultiStatistics
from moptipyapps.prodsched.rop_simulation import ROPSimulation
from moptipyapps.prodsched.statistics_collector import StatisticsCollector


class ROPMultiSimulation(Encoding):
    """A multi-simulation."""

    def __init__(self, instances: Iterable[Instance]) -> None:
        """
        Instantiate the multi-statistics decoding.

        :param instances: the packing instance
        """
        if not isinstance(instances, Iterable):
            raise type_error(instances, "instances", Iterable)
        #: the statistics collectors
        col: Final[tuple[StatisticsCollector, ...]] = tuple(
            StatisticsCollector(inst) for inst in instances)
        #: the simulations and collectors
        self.__simulations: Final[tuple[tuple[
            ROPSimulation, StatisticsCollector], ...]] = tuple(
            (ROPSimulation(inst, col[i]), col[i])
            for i, inst in enumerate(instances))

    def decode(self, x: np.ndarray, y: MultiStatistics) -> None:
        """
        Map a ROP setting to a multi-statistics.

        :param x: the array
        :param y: the Gantt chart
        """
        for i, (sim, col) in enumerate(self.__simulations):
            col.set_dest(y.per_instance[i])
            sim.ctrl_reset()
            sim.set_rop(x)
            sim.ctrl_run()

    def __str__(self) -> str:
        """
        Get the name of this decoding.

        :return: `"rms"`
        :rtype: str
        """
        return "rms"
